#!/usr/bin/python2.7
from numpy.core.multiarray import ndarray

from GT3.BeamDeposition.Functions.GetMFP import get_mfp, sigeff
from scipy.interpolate import UnivariateSpline
from scipy.integrate import quad, fixed_quad, cumtrapz
from math import sqrt, exp
from time import time
import warnings
import numpy as np
from scipy.optimize import root_scalar
from math import pi, isinf
import matplotlib.pyplot as plt
from collections import namedtuple
from scipy import constants
from GT3.Core.Functions.eVConvert import eVConvert

m_d = constants.physical_constants['deuteron mass'][0]

class Beam:

    beamE = None  # type: float
    beamP = None # type: float
    beamA = None # type: float
    beamWidth = None  # type: float
    shaf_shift_0 = None  # type: float
    dVdr = None  # type: UnivariateSpline
    dVdrho = None  # type: UnivariateSpline
    rtang = None  # type: float
    kappa_vals = None  # type: object
    r2vol = None  # type: UnivariateSpline
    timing = None  # type: bool
    rho = None  # type: np.ndarray
    verbose = None  # type: bool
    R0 = None # type: float
    Ti = None # type: np.ndarray
    Te = None # type: np.ndarray
    ne = None # type: np.ndarray
    iolFlag = True # type: bool
    coCurr = True # type: bool

    DepositionProfiles = namedtuple('DepositionProfiles', ['D1', 'D2', 'D3'])
    """A named tuple for deposition profiles for D1, D2 and D3 molecules"""
    EnergySplit = namedtuple('EnergySplit', ['D1', 'D2', 'D3'])
    PowerProfiles = namedtuple('PowerProfiles', ['D1', 'D2', 'D3', 'total'])
    MomentumProfiles = namedtuple('MomentumProfiles', ['D1', 'D2', 'D3', 'total', 'coCurr'])
    IOLSplit = namedtuple('IOLSplit', 'total lost kept')



    #def __init__(self, rho, r2vol, dVdr, rtang, shaf_shift, kappa_vals,
    #             beamWidth, Te, TC, R0, ne, beamE, beamA, beamP, Zeff, a, gaussR, verbose=False, timing=False):

    def __init__(self, config):
        """
        The Beams class for a single neutral beam injecter beam

        This class mimicks NBEAMS and provides the deposition profile H(rho,i) for the three molecular species.

        LIMITATIONS:

        This model assumes a circular beam with gaussian current density J. Beam shinethrough is not currently calculated
        correctly.

        Parameters
        ----------

        :param verbose: Enable verbose output
        :type verbose: bool
        :param timing: Enable timing data
        :type timing: bool
        :param rho: The rho vector
        :type rho:  np.ndarray
        :param r2vol: The r2vol interpolator
        :type r2vol:  UnivariateSpline
        :param dVdr: The dVdr interpolator
        :type dVdr:  UnivariateSpline
        :param rtang: The radius of tangency of this beam
        :type rtang:  float
        :param shaf_shift: The shavronof shift interpolator
        :type shaf_shift:  UnivariateSpline
        :param kappa_vals: The kappa values at the mag axes/sep
        :type kappa_vals:  np.ndarary(2)
        :param beamWidth: The beam width (2x the radius for circular)
        :type beamWidth:  float
        :param Te: The electron temperature np.ndarray
        :type Te:  np.ndarray
        :param TC:  The carbon temperature np.ndarray
        :type TC:  np.ndarray
        :param R0: The major radius at the magnetic axis
        :type R0:  float
        :param ne: The electron density np.ndarray
        :type ne:  np.ndarray
        :param beamE: The beam energy in keV
        :type beamE:  float
        :param beamA: The beam ion mass in u
        :type beamA:  float
        :param beamP: The beam power in MW (as scaled by the duty cycle)
        :type beamP: float
        :param Zeff:  The Z-effective np.array
        :type Zeff:  np.ndarray
        :param a: The plasma radius
        :type a:  float
        :param gaussR: The gaussian radius of hte circular beam
        :type gaussR:  float

        """

        self.config = config
        self.name = config.name
        start = time()
        self.verbose = config.verbose
        self.timing = config.timing
        self.coCurr = config.coCurr
        self.rho = config.rho
        self.iol = config.iol
        self.iolFlag = config.iolFlag

        self.r2vol = config.r2vol
        self.dVdr = config.dVdr
        self.dVdrho = config.dVdrho
        self.rtang = config.rtang
        self.shaf_shift_0 = config.shaf_shift
        self.kappa_vals = config.kappa_vals
        self.beamWidth = config.beamWidth
        self.Te = config.Te
        self.Tc = config.Ti
        self.R0 = config.R0
        self.ne = config.ne
        self.beamE = config.beamE * 1000
        self.beamA = config.beamA
        self.beamP = config.beamP
        self.Zeff = config.Z_eff
        self.a = config.a
        self.gaussR = config.gaussR
        self.quad = "fixed_quad"
        self.prof_root = {}
        self.prof_mfp = {}
        self.prof_D = {}
        self.kappa_vals = self.kappa_vals.axis + (self.kappa_vals.sep - self.kappa_vals.axis) * self.rho**2
        self.shift_vals = self.shaf_shift_0 * (1 - self.rho**2)
        self.shift = UnivariateSpline(self.rho, self.shift_vals)
        self.shift_prime = self.shift.derivative()
        self.kappa = UnivariateSpline(self.rho, self.kappa_vals)
        self.kappa_prime = self.kappa.derivative()
        self.vol = self.r2vol(self.a)
        self.pwrFracOverride = config.pwrFracOverride
        self.sep_val = self.rho[-1]

        self.prof_root['count'] = 0
        self.prof_root['time'] = []
        self.prof_mfp['count'] = 0
        self.prof_mfp['time'] = []
        self.prof_D['count'] = 0
        self.prof_D['time'] = []

        if not config.beamResult:
            self.is_new(config)
        else:
            old_result = config.beamResult
            try:

                self.angle_1 = UnivariateSpline(old_result[6], old_result[7], k=3, s=0)
                self.angle_2 = UnivariateSpline(old_result[6], old_result[8], k=3, s=0)
                self.angle_3 = UnivariateSpline(old_result[6], old_result[9], k=3, s=0)
                self.Hofrho = self.DepositionProfiles(UnivariateSpline(old_result[6], old_result[10], k=3, s=0),
                                                  UnivariateSpline(old_result[6], old_result[11], k=3, s=0),
                                                  UnivariateSpline(old_result[6], old_result[12], k=3, s=0))
                self.shine = old_result[13]
                self.Hofr = self.calc_Hofr(self.Hofrho, self.DepositionProfiles)
                """The normalized H(r) function on [0., a]"""
                if self.pwrFracOverride:
                    self.pwrfrac = self.pwrFracOverride
                    print "Power fraction overwritten: " + str(self.pwrfrac)
                else:
                    self.pwrfrac = self.calc_power_frac(self.beamE)
                self.dPdV = self.calc_dPdV(self.PowerProfiles)
                self.energies = self.EnergySplit(eVConvert(self.beamE), eVConvert(self.beamE / 2.),
                                            eVConvert(self.beamE / 3.))

                self.calc_iol(self.iol)
                self.calc_part_sources(self.IOLSplit)
                self.calc_heat_sources(self.IOLSplit)
                self.calc_mom_sources(self.IOLSplit)

            except Exception as e:
                print "Failed to rebuild beam %s with error: %s" % (self.name, str(e))


    def is_new(self, config):
        start = time()
        h_res_1 = None
        """The unnormalized H(rho) interpolator for Energy Group 1 on [0., 1.]"""
        h_res_2 = None
        """The unnormalized H(rho) interpolator for Energy Group 2 on [0., 1.]"""
        h_res_3 = None
        """The unnormalized H(rho) interpolator for Energy Group 3 on [0., 1.]"""

        self.angle_1 = None
        """The interpolator for Energy Group 1 for the launch angle on [0., 1.]"""
        self.angle_2 = None
        """The interpolator for Energy Group 2 for the launch angle on [0., 1.]"""
        self.angle_3 = None
        """The interpolator for Energy Group 3 for the launch angle on [0., 1.]"""

        print "Running %s Beam" % config.name
        if self.verbose: print "Calculating energy group 1"
        h_res_1, self.angle_1 = self.calcHofRho(self.beamE)

        if self.verbose: print  "Calculating energy group 2"
        h_res_2, self.angle_2 = self.calcHofRho(self.beamE / 2.)
        """The H(rho) and launch angle(rho) for Energy Group 2 on [0., 1.]"""

        if self.verbose: print "Calculating energy group 3"
        h_res_3, self.angle_3 = self.calcHofRho(self.beamE / 3.)  # type: (UnivariateSpline, UnivariateSpline)
        """The H(rho) and launch angle(rho) for Energy Group 3 on [0., 1.]"""


        self.Hofrho_unnorm = self.DepositionProfiles(h_res_1,
                                                h_res_2,
                                                h_res_3)

        self.shine = self.calc_shine()
        self.Hofrho = self.normalize_Hofrho(self.DepositionProfiles)
        self.Hofr = self.calc_Hofr(self.Hofrho, self.DepositionProfiles)
        """The normalized H(r) function on [0., a]"""
        if self.pwrFracOverride:
            self.pwrfrac = self.pwrFracOverride
            print "Power fraction overwritten: " + str(self.pwrfrac)
        else:
            self.pwrfrac = self.calc_power_frac(self.beamE)
        self.dPdV = self.calc_dPdV(self.PowerProfiles)
        self.energies = self.EnergySplit(eVConvert(self.beamE), eVConvert(self.beamE / 2.), eVConvert(self.beamE / 3.))
        self.calc_iol(self.iol)
        self.calc_part_sources(self.IOLSplit)
        self.calc_heat_sources(self.IOLSplit)
        self.calc_mom_sources(self.IOLSplit)


        end=time()

        if self.timing:
            print """
            Timing summary:
            
            Total runtime: %s seconds
            
            Number of rootfinding operations: %s
            Total time for rootfinding: %s seconds
            Number of MFP operations: %s
            Total time of MFP operations: %s seconds
            Number of calculations of D: %s
            Total time of D calculations: %s seconds""" % (end - start,
                                                           self.prof_root['count'],
                                                           np.sum(self.prof_root['time']),
                                                           self.prof_mfp['count'],
                                                           np.sum(self.prof_mfp['time']),
                                                           self.prof_D['count'],
                                                           np.sum(self.prof_D['time']))


    def calc_iol(self, iol):

        """
        Calculate the Ion Orbit Loss coefficients from the given iol module results. Returns nothing.
        The momentum loss fraction is retuned as the absolute value since there shouldn't realistically be any direction
        change.

        :param iol:
        """
        # TODO: Introduce accurate launch angles
        if iol:
            D1_iol = iol.calc_iol_beams(1, m_d, iol.iol_p, 33,  sqrt(2 * self.beamE * 1.6021E-19 / m_d), -.96, iol.coslist)
            D2_iol = iol.calc_iol_beams(1, m_d, iol.iol_p, 33, sqrt(2 * self.beamE * 1.6021E-19 / (m_d * 2.)), -.96,
                                        iol.coslist)
            D3_iol = iol.calc_iol_beams(1, m_d, iol.iol_p, 33, sqrt(2 * self.beamE * 1.6021E-19 / (3. * m_d)), -.96,
                                        iol.coslist)
            # Since we don't know if the beam meshing will be the same as the iol rho meshing, we map by index

            self.F_orb_D1 = np.zeros(len(self.rho))
            self.M_orb_D1 = np.zeros(len(self.rho))
            self.E_orb_D1 = np.zeros(len(self.rho))
            self.F_orb_D1[np.where(self.rho > iol.rho[:, 0][np.where(D1_iol[0][:, 0] > .1)[0].min()])[0].min():] = 0.5
            self.M_orb_D1[np.where(self.rho > iol.rho[:, 0][np.where(np.abs(D1_iol[1][:, 0]) > .1)[0].min()])[0].min():] = 0.5
            self.E_orb_D1[np.where(self.rho > iol.rho[:, 0][np.where(D1_iol[2][:, 0] > .1)[0].min()])[0].min():] = 0.5

            self.F_orb_D2 = np.zeros(len(self.rho))
            self.M_orb_D2 = np.zeros(len(self.rho))
            self.E_orb_D2 = np.zeros(len(self.rho))
            self.F_orb_D2[np.where(self.rho > iol.rho[:, 0][np.where(D2_iol[0][:, 0] > .1)[0].min()])[0].min():] = 0.5
            self.M_orb_D2[np.where(self.rho > iol.rho[:, 0][np.where(np.abs(D2_iol[1][:, 0]) > .1)[0].min()])[0].min():] = 0.5
            self.E_orb_D2[np.where(self.rho > iol.rho[:, 0][np.where(D2_iol[2][:, 0] > .1)[0].min()])[0].min():] = 0.5

            self.F_orb_D3 = np.zeros(len(self.rho))
            self.M_orb_D3 = np.zeros(len(self.rho))
            self.E_orb_D3 = np.zeros(len(self.rho))
            self.F_orb_D3[np.where(self.rho > iol.rho[:, 0][np.where(D3_iol[0][:,0]>.1)[0].min()])[0].min():] = 0.5
            self.M_orb_D3[np.where(self.rho > iol.rho[:, 0][np.where(np.abs(D3_iol[1][:, 0]) > .1)[0].min()])[0].min():] = 0.5
            self.E_orb_D3[np.where(self.rho > iol.rho[:, 0][np.where(D3_iol[2][:, 0] > .1)[0].min()])[0].min():] = 0.5

        else:
            self.F_orb_D1 = np.zeros(len(self.rho))
            self.F_orb_D2 = np.zeros(len(self.rho))
            self.F_orb_D3 = np.zeros(len(self.rho))

            self.E_orb_D1 = np.zeros(len(self.rho))
            self.E_orb_D2 = np.zeros(len(self.rho))
            self.E_orb_D3 = np.zeros(len(self.rho))

            self.M_orb_D1 = np.zeros(len(self.rho))
            self.M_orb_D2 = np.zeros(len(self.rho))
            self.M_orb_D3 = np.zeros(len(self.rho))




    def calc_part_sources(self, IOLSplit):
        """
        Calculate the particle sources in #/m^3*s as well as dS/d(rho)
        """
        self.part_src_dens_D1 = None  # type: ndarray
        """The particle source density as a function of r in #/(m^3 s) for the D1 beam"""
        self.part_src_dens_D2 = None  # type: ndarray
        """The particle source density as a function of r in #/(m^3 s) for the D2 beam"""
        self.part_src_dens_D3 = None  # type: ndarray
        """The particle source density as a function of r in #/(m^3 s) for the D3 beam"""
        self.part_src_D1 = None  #type: ndarray
        """The particle source as a function of r in #/(s) for the D1 beam"""
        self.part_src_D2 = None  #type: ndarray
        """The particle source as a function of r in #/(s) for the D2 beam"""
        self.part_src_D2 = None  #type: ndarray
        """The particle source as a function of r in #/(s) for the D3 beam"""

        if self.iolFlag:

            self.part_src_dens_D1 = IOLSplit(self.dPdV.D1(self.rho) / self.energies.D1.J,  # total
                                             self.F_orb_D1 * self.dPdV.D1(self.rho) / self.energies.D1.J,  # lost
                                             (1. - self.F_orb_D1) * self.dPdV.D1(self.rho) / self.energies.D1.J)  # kept

            self.part_src_dens_D2 = IOLSplit(self.dPdV.D2(self.rho) / self.energies.D2.J,  # total
                                             self.F_orb_D2 * self.dPdV.D2(self.rho) / self.energies.D2.J,  # lost
                                             (1. - self.F_orb_D2) * self.dPdV.D2(self.rho) / self.energies.D2.J)  # kept

            self.part_src_dens_D3 = IOLSplit(self.dPdV.D3(self.rho) / self.energies.D3.J,  # total
                                        self.F_orb_D3 * self.dPdV.D3(self.rho) / self.energies.D3.J,  # lost
                                        (1. - self.F_orb_D3) * self.dPdV.D3(self.rho) / self.energies.D3.J)  # kept

            self.part_src_D1 = IOLSplit(self.dPdV.D1(self.rho) * self.dVdrho(self.rho) / self.energies.D1.J,  # total
                                        self.F_orb_D1 * self.dPdV.D1(self.rho) * self.dVdrho(self.rho) / self.energies.D1.J,  # lost
                                        (1. - self.F_orb_D1) * self.dPdV.D1(self.rho) * self.dVdrho(self.rho) / self.energies.D1.J)  # kept

            self.part_src_D2= IOLSplit(self.dPdV.D2(self.rho) * self.dVdrho(self.rho) / self.energies.D2.J,  # total
                                        self.F_orb_D2 * self.dPdV.D2(self.rho) * self.dVdrho(self.rho) / self.energies.D2.J,  # lost
                                        (1. - self.F_orb_D2) * self.dPdV.D2(self.rho) * self.dVdrho(self.rho) / self.energies.D2.J)  # kept

            self.part_src_D3 = IOLSplit(self.dPdV.D3(self.rho) * self.dVdrho(self.rho) / self.energies.D3.J,  # total
                                        self.F_orb_D3 * self.dPdV.D3(self.rho) * self.dVdrho(self.rho) / self.energies.D3.J,  # lost
                                        (1. - self.F_orb_D3) * self.dPdV.D3(self.rho) * self.dVdrho(self.rho) / self.energies.D3.J)  # kept


        else:
            self.part_src_dens_D1 = IOLSplit(self.dPdV.D1(self.rho) / self.energies.D1.J,
                                        np.zeros(len(self.rho)),
                                        self.dPdV.D1(self.rho) / self.energies.D1.J)

            self.part_src_dens_D2 = IOLSplit(self.dPdV.D1(self.rho) / self.energies.D2.J,
                                        np.zeros(len(self.rho)),
                                        self.dPdV.D1(self.rho) / self.energies.D2.J)

            self.part_src_dens_D3 = IOLSplit(self.dPdV.D1(self.rho) / self.energies.D3.J,
                                        np.zeros(len(self.rho)),
                                        self.dPdV.D1(self.rho) / self.energies.D3.J)

            self.part_src_D1 = IOLSplit(self.dPdV.D1(self.rho) * self.dVdrho(self.rho) / self.energies.D1.J,  # total
                                        self.dPdV.D1(self.rho) * self.dVdrho(self.rho) / self.energies.D1.J,  # lost
                                        self.dPdV.D1(self.rho) * self.dVdrho(self.rho) / self.energies.D1.J)  # kept

            self.part_src_D2 = IOLSplit(self.dPdV.D2(self.rho) * self.dVdrho(self.rho) / self.energies.D2.J,  # total
                                        self.dPdV.D2(self.rho) * self.dVdrho(self.rho) / self.energies.D2.J,  # lost
                                        self.dPdV.D2(self.rho) * self.dVdrho(self.rho) / self.energies.D2.J)  # kept

            self.part_src_D3 = IOLSplit(self.dPdV.D3(self.rho) * self.dVdrho(self.rho) / self.energies.D3.J,  # total
                                        self.dPdV.D3(self.rho) * self.dVdrho(self.rho) / self.energies.D3.J,  # lost
                                        self.dPdV.D3(self.rho) * self.dVdrho(self.rho) / self.energies.D3.J)  # kept


        self.part_src_dens = IOLSplit(self.part_src_dens_D1.total + self.part_src_dens_D2.total + self.part_src_dens_D3.total,
                                      self.part_src_dens_D1.lost + self.part_src_dens_D2.lost + self.part_src_dens_D3.lost,
                                      self.part_src_dens_D1.kept + self.part_src_dens_D2.kept + self.part_src_dens_D3.kept)
        """The particle source density IOL split as a function of r in #/(m^3 s) """

        self.part_src = IOLSplit(self.part_src_D1.total + self.part_src_D2.total + self.part_src_D3.total,
                                 self.part_src_D1.lost + self.part_src_D2.lost + self.part_src_D3.lost,
                                 self.part_src_D1.kept + self.part_src_D2.kept + self.part_src_D3.kept)
        """The particle source IOL split as a function of r in #/s """

        # Remove any values that are stupidly small but not 0.
        #
        # self.part_src_tot_total[self.part_src_tot_total < 1E8] = 0.0
        # self.part_src_tot_kept[self.part_src_tot_kept < 1E8] = 0.0
        # self.part_src_tot_lost[self.part_src_tot_lost < 1E8] = 0.0
        # self.part_src_tot_of_rho[self.part_src_tot_of_rho < 1E8] = 0.0


    def calc_heat_sources(self, IOLSplit):

        self.en_src_dens_D1 = None  # type: ndarray
        """The energy source density as a function of r in #/(m^3 s) for the D1 beam"""
        self.en_src_dens_D2 = None  # type: ndarray
        """The energy source density as a function of r in #/(m^3 s) for the D2 beam"""
        self.en_src_dens_D3 = None  # type: ndarray
        """The energy source density as a function of r in #/(m^3 s) for the D3 beam"""
        self.en_src_D1 = None  #type: ndarray
        """The energy source as a function of r in #/(s) for the D1 beam"""
        self.en_src_D2 = None  #type: ndarray
        """The energy source as a function of r in #/(s) for the D2 beam"""
        self.en_src_D2 = None  #type: ndarray
        """The energy source as a function of r in #/(s) for the D3 beam"""

        if self.iolFlag:
            self.en_src_dens_D1 = IOLSplit(self.dPdV.D1(self.rho),  # total
                                      self.E_orb_D1 * self.dPdV.D1(self.rho),  # lost
                                      (1. - self.E_orb_D1) * self.dPdV.D1(self.rho))   # kept

            self.en_src_dens_D2 = IOLSplit(self.dPdV.D2(self.rho),  # total
                                      self.E_orb_D2 * self.dPdV.D2(self.rho),  # lost
                                      (1. - self.E_orb_D2) * self.dPdV.D2(self.rho))   # kept

            self.en_src_dens_D3 = IOLSplit(self.dPdV.D3(self.rho),  # total
                                      self.E_orb_D3 * self.dPdV.D3(self.rho),  # lost
                                      (1. - self.E_orb_D3) * self.dPdV.D3(self.rho))   # kept

            self.en_src_D1 = IOLSplit(self.dPdV.D1(self.rho) * self.dVdrho(self.rho),  # total
                                        self.E_orb_D1 * self.dPdV.D1(self.rho) * self.dVdrho(self.rho),  # lost
                                        (1. - self.E_orb_D1) * self.dPdV.D1(self.rho) * self.dVdrho(self.rho))  # kept

            self.en_src_D2 = IOLSplit(self.dPdV.D2(self.rho) * self.dVdrho(self.rho),  # total
                                        self.E_orb_D2 * self.dPdV.D2(self.rho) * self.dVdrho(self.rho),  # lost
                                        (1. - self.E_orb_D2) * self.dPdV.D2(self.rho) * self.dVdrho(self.rho))  # kept

            self.en_src_D3 = IOLSplit(self.dPdV.D3(self.rho) * self.dVdrho(self.rho),  # total
                                        self.E_orb_D3 * self.dPdV.D3(self.rho) * self.dVdrho(self.rho) ,  # lost
                                        (1. - self.E_orb_D3) * self.dPdV.D3(self.rho) * self.dVdrho(self.rho))  # kept

        else:
            self.en_src_dens_D1 = IOLSplit(self.dPdV.D1(self.rho),  # total
                                      np.zeros(len(self.rho)),  # lost
                                      self.dPdV.D1(self.rho))   # kept

            self.en_src_dens_D2 = IOLSplit(self.dPdV.D2(self.rho),  # total
                                      np.zeros(len(self.rho)),  # lost
                                      self.dPdV.D2(self.rho))   # kept

            self.en_src_dens_D3 = IOLSplit(self.dPdV.D3(self.rho),  # total
                                      np.zeros(len(self.rho)),  # lost
                                      self.dPdV.D3(self.rho))   # kept

            self.en_src_D1 = IOLSplit(self.dPdV.D1(self.rho) * self.dVdrho(self.rho),  # total
                                        self.dPdV.D1(self.rho) * self.dVdrho(self.rho),  # lost
                                        self.dPdV.D1(self.rho) * self.dVdrho(self.rho))  # kept

            self.en_src_D2 = IOLSplit(self.dPdV.D2(self.rho) * self.dVdrho(self.rho),  # total
                                        self.dPdV.D2(self.rho) * self.dVdrho(self.rho),  # lost
                                        self.dPdV.D2(self.rho) * self.dVdrho(self.rho))  # kept

            self.en_src_D3 = IOLSplit(self.dPdV.D3(self.rho) * self.dVdrho(self.rho),  # total
                                        self.dPdV.D3(self.rho) * self.dVdrho(self.rho),  # lost
                                        self.dPdV.D3(self.rho) * self.dVdrho(self.rho))  # kept


        self.en_src_dens = IOLSplit(self.en_src_dens_D1.total + self.en_src_dens_D2.total + self.en_src_dens_D3.total,
                                      self.en_src_dens_D1.lost + self.en_src_dens_D2.lost + self.en_src_dens_D3.lost,
                                      self.en_src_dens_D1.kept + self.en_src_dens_D2.kept + self.en_src_dens_D3.kept)
        """The particle source density IOL split as a function of r in J/(m^3 s) """

        self.en_src = IOLSplit(self.en_src_D1.total + self.en_src_D2.total + self.en_src_D3.total,
                                 self.en_src_D1.lost + self.en_src_D2.lost + self.en_src_D3.lost,
                                 self.en_src_D1.kept + self.en_src_D2.kept + self.en_src_D3.kept)
        """The particle source IOL split as a function of r in #/s """

        # Zero out very small values
        #
        # self.en_src_tot_total[self.en_src_tot_total < 1E2] = 0.0
        # self.en_src_tot_kept[self.en_src_tot_kept < 1E2] = 0.0
        # self.en_src_tot_lost[self.en_src_tot_lost < 1E2] = 0.0
        # self.en_src_tot_of_rho[self.en_src_tot_of_rho < 1E2] = 0.0


    def calc_mom_sources(self, IOLSplit):

        if self.iolFlag:
            self.mom_src_dens_D1 = IOLSplit(self.dPdV.D1(self.rho) * np.sqrt(0.5 * m_d / self.energies.D1.J) * self.rtang * self.angle_1(self.rho),
                                       self.M_orb_D1 * self.dPdV.D1(self.rho)  * np.sqrt(0.5 * m_d / self.energies.D1.J) * self.rtang * self.angle_1(self.rho),
                                       (1. - self.M_orb_D1) * self.dPdV.D1(self.rho) * np.sqrt(0.5 * m_d / self.energies.D1.J) * self.rtang * self.angle_1(self.rho))

            self.mom_src_dens_D2 = IOLSplit(self.dPdV.D2(self.rho) * np.sqrt(0.5 * m_d / self.energies.D2.J) * self.rtang * self.angle_2(self.rho),
                                       self.M_orb_D2 * self.dPdV.D2(self.rho)  * np.sqrt(0.5 * m_d / self.energies.D2.J) * self.rtang * self.angle_2(self.rho),
                                       (1. - self.M_orb_D2) * self.dPdV.D2(self.rho)  * np.sqrt(0.5 * m_d / self.energies.D2.J) * self.rtang * self.angle_2(self.rho))

            self.mom_src_dens_D3 = IOLSplit(self.dPdV.D3(self.rho) * np.sqrt(0.5 * m_d / self.energies.D3.J) * self.rtang * self.angle_3(self.rho),
                                       self.M_orb_D3 * self.dPdV.D3(self.rho)  * np.sqrt(0.5 * m_d / self.energies.D3.J) * self.rtang * self.angle_3(self.rho),
                                       (1. - self.M_orb_D3) * self.dPdV.D3(self.rho)  * np.sqrt(0.5 * m_d / self.energies.D3.J) * self.rtang * self.angle_3(self.rho))

            self.mom_src_D1 = IOLSplit(self.dPdV.D1(self.rho) * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D1.J) * self.rtang * self.angle_1(self.rho),
                                       self.M_orb_D1 * self.dPdV.D1(self.rho) * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D1.J) * self.rtang * self.angle_1(self.rho),
                                       (1. - self.M_orb_D1) * self.dPdV.D1(self.rho) * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D1.J) * self.rtang * self.angle_1(self.rho))

            self.mom_src_D2 = IOLSplit(self.dPdV.D2(self.rho) * np.sqrt(0.5 * m_d / self.energies.D2.J) * self.rtang * self.angle_2(self.rho),
                                       self.M_orb_D2 * self.dPdV.D2(self.rho) * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D2.J) * self.rtang * self.angle_2(self.rho),
                                       (1. - self.M_orb_D2) * self.dPdV.D2(self.rho) * np.sqrt(0.5 * m_d / self.energies.D2.J) * self.rtang * self.angle_2(self.rho))

            self.mom_src_D3 = IOLSplit(self.dPdV.D3(self.rho) * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D3.J) * self.rtang * self.angle_3(self.rho),
                                       self.M_orb_D3 * self.dPdV.D3(self.rho)  * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D3.J) * self.rtang * self.angle_3(self.rho),
                                       (1. - self.M_orb_D3) * self.dPdV.D3(self.rho)  * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D3.J) * self.rtang * self.angle_3(self.rho))


            if not self.coCurr:
                self.mom_src_D1 = -1. * self.mom_src_D1
                self.mom_src_D2 = -1. * self.mom_src_D2
                self.mom_src_D3 = -1. * self.mom_src_D3
                self.mom_src_D1_of_rho = -1 * self.mom_src_D1_of_rho
                self.mom_src_D2_of_rho = -1 * self.mom_src_D2_of_rho
                self.mom_src_D3_of_rho = -1 * self.mom_src_D3_of_rho

        else:
            self.mom_src_dens_D1 = IOLSplit(self.dPdV.D1(self.rho) * np.sqrt(0.5 * m_d / self.energies.D1.J) * self.rtang * self.angle_1,
                                       np.zeros(len(self.rho)),
                                       self.dPdV.D1(self.rho) * np.sqrt(0.5 * m_d / self.energies.D1.J) * self.rtang * self.angle_1)

            self.mom_src_dens_D2 = IOLSplit(self.dPdV.D2(self.rho) * np.sqrt(0.5 * m_d / self.energies.D2.J) * self.rtang * self.angle_2,
                                       np.zeros(len(self.rho)),
                                       self.dPdV.D2(self.rho) * np.sqrt(0.5 * m_d / self.energies.D2.J) * self.rtang * self.angle_2)

            self.mom_src_dens_D3 = IOLSplit(self.dPdV.D3(self.rho) * np.sqrt(0.5 * m_d / self.energies.D3.J) * self.rtang * self.angle_3,
                                       np.zeros(len(self.rho)),
                                       self.dPdV.D3 * np.sqrt(0.5 * m_d / self.energies.D3.J) * self.rtang * self.angle_3)

            self.mom_src_D1 = IOLSplit(self.dPdV.D1(self.rho) * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D1.J) * self.rtang * self.angle_1,
                                       np.zeros(len(self.rho)),
                                       self.dPdV.D1(self.rho) * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D1.J) * self.rtang * self.angle_1)

            self.mom_src_D2 = IOLSplit(self.dPdV.D2(self.rho) * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D2.J) * self.rtang * self.angle_2,
                                       np.zeros(len(self.rho)),
                                       self.dPdV.D2(self.rho) * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D2.J) * self.rtang * self.angle_2)

            self.mom_src_D3 = IOLSplit(self.dPdV.D3(self.rho) * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D3.J) * self.rtang * self.angle_3,
                                       np.zeros(len(self.rho)),
                                       self.dPdV.D3(self.rho) * self.dVdrho(self.rho) * np.sqrt(0.5 * m_d / self.energies.D3.J) * self.rtang * self.angle_3)
            if not self.coCurr:
                self.mom_src_dens_D1 = -1. * self.mom_src_D1
                self.mom_src_dens_D2 = -1. * self.mom_src_D2
                self.mom_src_dens_D3 = -1. * self.mom_src_D3
                self.mom_src_D1 = -1 * self.mom_src_D1_of_rho
                self.mom_src_D2 = -1 * self.mom_src_D2_of_rho
                self.mom_src_D3 = -1 * self.mom_src_D3_of_rho

        self.mom_src_dens = IOLSplit(self.mom_src_dens_D1.total + self.mom_src_dens_D2.total + self.mom_src_dens_D3.total,
                                      self.mom_src_dens_D1.lost + self.mom_src_dens_D2.lost + self.mom_src_dens_D3.lost,
                                      self.mom_src_dens_D1.kept + self.mom_src_dens_D2.kept + self.mom_src_dens_D3.kept)
        """The momentum source density IOL split as a function of r in #/(m^3 s) """

        self.mom_src = IOLSplit(self.mom_src_D1.total + self.mom_src_D2.total + self.mom_src_D3.total,
                                 self.mom_src_D1.lost + self.mom_src_D2.lost + self.mom_src_D3.lost,
                                 self.mom_src_D1.kept + self.mom_src_D2.kept + self.mom_src_D3.kept)
        """The momentum source IOL split as a function of r in #/s """


    def calc_dPdV(self, Profile):
        """

        :param Profile:
        :return: Univariate splines of D1, D2, D3, and the total
        :rtype: Profile(UnivariateSpline)
        """
        y1 = lambda x: self.Hofrho.D1(x) * self.beamP *self.pwrfrac[0] * 1E6 / self.vol
        y2 = lambda x: self.Hofrho.D2(x) * self.beamP *self.pwrfrac[1] * 1E6 / self.vol
        y3 = lambda x: self.Hofrho.D3(x) * self.beamP *self.pwrfrac[2] * 1E6 / self.vol
        y4 = lambda x: y1(x) + y2(x) + y3(x)

        interp_1 = UnivariateSpline(self.rho, y1(self.rho), k=3, s=2)
        """The dP(rho)/dV"""
        interp_2 = UnivariateSpline(self.rho, y2(self.rho), k=3, s=2)
        interp_3 = UnivariateSpline(self.rho, y3(self.rho), k=3, s=2)
        interp_4 = UnivariateSpline(self.rho, y4(self.rho), k=3, s=2)


        return Profile(interp_1,
                       interp_2,
                       interp_3,
                       interp_4)

    def calc_power_frac(self, beamEeV):
        """
        Calculate the D3D power fraction

        TODO: Accept arbitrary functions into the Beams class to calculate this.

        :param beamE: The beam power
        :type beamE: float
        :return: A list of the D1/D2/D3 power fracs
        :rtype: list
        """
        beamE = beamEeV / 1E3
        pwr_frac = np.zeros(3)
        pwr_frac[0] = (68 + 0.11 * beamE) / 100
        pwr_frac[1] = (-159 + 6.53 * beamE - 0.082 * (beamE ** 2) + 0.00034 * (beamE ** 3)) / 100
        pwr_frac[2] = (191 - 6.64 * beamE + 0.082 * (beamE ** 2) - 0.00034 * (beamE ** 3)) / 100
        return pwr_frac

    def calc_shine(self):
        shine_func_1 = lambda x: self.dVdr(x * self.a) * self.Hofrho_unnorm.D1(x)
        shine_func_2 = lambda x: self.dVdr(x * self.a) * self.Hofrho_unnorm.D2(x)
        shine_func_3 = lambda x: self.dVdr(x * self.a) * self.Hofrho_unnorm.D3(x)

        shine = 1 - (cumtrapz(shine_func_1(self.rho), self.rho)[-1] / self.vol +
                          cumtrapz(shine_func_2(self.rho), self.rho)[-1] / self.vol +
                          cumtrapz(shine_func_3(self.rho), self.rho)[-1] / self.vol)

        if shine <= 0. or shine > 1.:

             warnings.warn("Shinethrough is out of range at %s" % shine)

        return shine

    def normalize_Hofrho(self, DepClass):
        """
        Normalize H(rho) such that integral(H(rho)dV) = Vplasma and provide an interpolator
        Note that to do this integration from 0->1, you must multiply the integrand by dV/d(rho)

        :type DepClass:
        """

        norm_int_D1 = lambda x: self.dVdr(x * self.a) * self.Hofrho_unnorm.D1(x) * self.a
        norm_int_D2 = lambda x: self.dVdr(x * self.a) * self.Hofrho_unnorm.D2(x) * self.a
        norm_int_D3 = lambda x: self.dVdr(x * self.a) * self.Hofrho_unnorm.D3(x) * self.a

        normal_const_D1 = self.vol / cumtrapz(norm_int_D1(self.rho), self.rho)[-1]
        normal_const_D2 = self.vol / cumtrapz(norm_int_D2(self.rho), self.rho)[-1]
        normal_const_D3 = self.vol / cumtrapz(norm_int_D3(self.rho), self.rho)[-1]

        interp_1 = UnivariateSpline(self.rho, self.Hofrho_unnorm.D1(self.rho) * normal_const_D1)
        interp_2 = UnivariateSpline(self.rho, self.Hofrho_unnorm.D2(self.rho) * normal_const_D2)
        interp_3 = UnivariateSpline(self.rho, self.Hofrho_unnorm.D3(self.rho) * normal_const_D3)

        return DepClass(interp_1,
                        interp_2,
                        interp_3)

    def calc_Hofr(self, Hofrho, depClass):
        """
        Calculate H(r) and return the depClass named tuple containing 3 interpolators as functions of r

        :param depClass: The deposition class object to be used
        :param Hofrho: The H(rho) function
        :type Hofrho: DepositionProfiles
        :return: The H(r) np.ndarray
        :rtype: depClass
        """
        y1 = lambda x: Hofrho.D1(x * self.a)
        y2 = lambda x: Hofrho.D2(x * self.a)
        y3 = lambda x: Hofrho.D3(x * self.a)
        interp_space = np.linspace(0., self.a, len(self.rho))

        return depClass(UnivariateSpline(interp_space, y1(self.rho), k=3, s=0),
                             UnivariateSpline(interp_space, y2(self.rho), k=3, s=0),
                             UnivariateSpline(interp_space, y3(self.rho), k=3, s=0))

    def calcHofRho(self, energy):
        """

        Calculate the beam deposition profile

        'quad' - scipy.integrate.quad - Uses Fortran QUADPACK technique
        'fixed_quad' - scipy.integrate.fixed_quad - Fixed-order gaussian quadrature (n=5)
        'trap' - scipy.integrate.cumtrapz - Composite trapazoidal rule integration (n=10)

        Fixed quadrature seems to get the same answer on a few tests in a fraction of the time

        TODO: Run more extensive tests to validate quadrature choice
        :param energy: The energy group (in keV)
        :return: UnivariateSplines of H(rho) and ksi(rho) interpolated on [0., 1.]
        :rtype: list(UnivariateSpline)

        """

        # Circular beam

        result_pos = np.zeros(len(self.rho))

        result_minus = np.zeros(len(self.rho))

        Hresult = np.zeros(len(self.rho))
        angle_result = np.zeros(len(self.rho))
        mfp = get_mfp(self.ne, energy, self.Te, self.beamA, self.Zeff, self.rho)
        smallFlag = False

        for n, val in enumerate(self.rho):
        #for n, val in enumerate([.5]):
            if self.verbose: print val
            if float(val)== self.sep_val:
                Hresult[n] = 0.
                angle_result[n] = 0.
                continue
            if val <=0.05:
                smallFlag = True
            else:
                smallFlag = False
            posFlag = True
            Zulimit= min(val * self.a * self.kappa(val), self.beamWidth/2.)
            RLowerLimit = lambda x: self.rtang - sqrt((self.beamWidth/2.)**2 - x**2)
            Rpm = lambda x, n=val: self.get_Rpm(x, n, posFlag)
            hofrpzJLambda = lambda y, x, val=val: self.hofrpzJ(x, y, val, mfp, posFlag, smallFlag)
            angleFunTopLambda = lambda y, x, val=val: self.hofrpzJ(x, y, val, mfp, posFlag, smallFlag) * y / Rpm(x)

            RUpperLimit = lambda x: min(Rpm(x), self.rtang + sqrt((self.beamWidth / 2.0)**2 - x**2))
            if self.verbose: print "Beginning integration of node n=%s" % n
            now = time()

            if self.verbose: print "Running positive"
            if smallFlag:
                result_pos[n] = pi * self.kappa(val) * self.beamdblquad(hofrpzJLambda, 0., Zulimit, RLowerLimit, RUpperLimit, val)
            else:
                result_pos[n] = 2.0 * self.beamdblquad(hofrpzJLambda, 0., Zulimit, RLowerLimit, RUpperLimit, val)
            posFlag = False

            if self.verbose: print "Running negative"
            if smallFlag:
                result_minus[n] = pi * self.kappa(val) * self.beamdblquad(hofrpzJLambda, 0., Zulimit, RLowerLimit, RUpperLimit,
                                                                        val)
            else:
                result_minus[n] = 2.0 * self.beamdblquad(hofrpzJLambda, 0.0, Zulimit, RLowerLimit, RUpperLimit, val)
            total = time() - now
            Hresult[n] = result_pos[n] + result_minus[n]

            posFlag = True
            angle_top_pos = self.beamdblquad(angleFunTopLambda, 0., Zulimit, RLowerLimit, RUpperLimit, val)
            angle_bottom_pos = result_pos[n] / 2.
            posFlag = False
            angle_top_neg = self.beamdblquad(angleFunTopLambda, 0., Zulimit, RLowerLimit, RUpperLimit, val)
            angle_bottom_neg = result_minus[n] / 2.

            # TODO: Still don't know which is correct

            angle_result[n] = (angle_top_pos + angle_top_neg)/(angle_bottom_neg + angle_bottom_pos)
            #angle_result[n] = (angle_top_pos / angle_bottom_pos) + (angle_top_neg / angle_bottom_neg)

            if self.verbose: print "Node n=%s completed in %s s" % (n, total)
            if self.verbose: print "Result: %s" % Hresult[n]
            if self.timing: print "Root scalar count: %s" % self.prof_root["count"]
            if self.timing: print "Root scalar total time: %s" % np.sum(self.prof_root['time'])
            if self.timing: print "MFP count: %s" % self.prof_mfp["count"]
            if self.timing: print "MFP total time: %s" % np.sum(self.prof_mfp['time'])
            if self.timing: print "D calc count: %s" % self.prof_D["count"]
            if self.timing: print "D calc total time: %s" % np.sum(self.prof_D['time'])

        Hofrho=UnivariateSpline(self.rho, Hresult, k=3, s=0)
        """The H(rho) interpolator on [0., 1.]"""
        angle = UnivariateSpline(self.rho, np.nan_to_num(angle_result), k=3, s=0)
        """The launch angle as a function of rho interpolator on [0., 1.]"""
        return Hofrho, angle

    def beamdblquad(self, f, yllimit, yulimit, xllimit, xulimit, *args):
        """
        Custom double integration function for f(Rb, Zb) integration.

        Integration limits are Z -> [yllimit, yulimit], R->[xllimit, xulimit]

        :rtype: float
        :return: Returns the numerical result of this integration
        :param f: The function to be ingrated in R,Z space
        :type f: function
        :param yllimit: The Zb lower limit
        :type yllimit: float
        :param yulimit: The Zb upper limit
        :type yulimit: float
        :param xllimit: The Rb lower limit
        :type xllimit: float
        :param xulimit: The Rb upper limit
        :type xulimit: float
        :param args:
        """
        if yllimit == yulimit:
            return 0
        rho_val = args[0]
        yres = 10
        mesh = np.linspace(yllimit, yulimit, yres)
        results = np.zeros(yres)

        for a in range(0, yres):
            flambda = lambda x, y=mesh[a], val=rho_val: f(x, y, val)
            if xulimit(mesh[a]) < xllimit(mesh[a]):
                results[a] = 0.
            elif (xulimit(mesh[a]) - xllimit(mesh[a])) / xllimit(mesh[a]) <= 10**(-5):
                results[a] = 0.
            elif a == yres-1:
                results[a] = results[a-1]
            else:
                try:
                    results[a] = fixed_quad(flambda, xllimit(mesh[a]), xulimit(mesh[a]), n=4)[0] * (yulimit - yllimit) / yres
                except:
                    print "Something went wrong"
                if self.verbose: print "Integration result for a = %s: %s" % (a, results[a])
        return np.sum(results)

    def hofrpzJ(self, Zb, Rb, rho_val, mfp, p, smallFlag):
        """

        :param smallFlag: If rho is small, the Z term will integrate to pi*k/2., which is handle in the main loop
        :param rho_val:  The rho value
        :type rho_val:  float
        :param p: Whether this is for positive (True) or negative (False)
        :type p:  bool
        :return:
        :param Zb: The beam-Z coordinate
        :type Zb: float
        :param Rb:
        :type Rb: THe beam-R coordinate
        :return:
        """
        if self.verbose: print "Coordinates: (Rb,Z) = (%s, %s)" % (Rb, Zb)
        Rpm_of_Z = lambda x, n=rho_val: self.get_Rpm(x, n, p)

        RLowerLimit = lambda x: self.rtang - sqrt((self.beamWidth / 2.) ** 2 - x ** 2)

        RUpperLimit = lambda x: min(Rpm_of_Z(x), self.rtang + sqrt((self.beamWidth / 2.0) ** 2 - x ** 2))

        if type(Rb) == np.ndarray:
            result = []
            for a in Rb:
                r = sqrt(Zb ** 2 + (self.rtang - a) ** 2)
                if (self.beamWidth / 2.0) ** 2 - Zb ** 2 <= 0:
                    result.append(0.)
                else:
                    if RUpperLimit(Zb):
                        if RLowerLimit(Zb) > RUpperLimit(Zb):
                            result.append(0.)
                        elif Rpm_of_Z(Zb) <= (self.rtang - sqrt((rho_val*self.a)**2 - ((Zb**2)/(self.kappa(rho_val)**2)))):
                            result.append(0.)
                        elif r > self.beamWidth / 2.:
                            result.append(0.)
                        else:
                        # print "Running for (Zb, Rb, x) = (%s, %s, %s)" % (Zb, Rb, x)
                            result.append(self.hofrpz(a, Zb, rho_val, mfp, p, smallFlag) * self.J(a, Zb))

                    else:
                        result.append(0.)
            return result
        else:
            r = sqrt(Zb ** 2 + (self.rtang - Rb) ** 2)
            if (self.beamWidth / 2.0) ** 2 - Zb ** 2 <= 0:
                return 0
            if RUpperLimit(Zb):
                if r > self.beamWidth / 2.:
                    return 0
                if RLowerLimit(Zb) > RUpperLimit(Zb):
                    return 0
                elif Rpm_of_Z(Zb) <= (self.rtang - sqrt((rho_val*self.a) ** 2 - ((Zb ** 2) / (self.kappa(rho_val) ** 2)))):
                    return 0
                return self.hofrpz(Rb, Zb, rho_val, mfp, p, smallFlag) * self.J(Rb, Zb)
            else:
                return 0.

    def J(self, Rb, Zb):
        """
        The beam current function at a given Rb,Zb coordinate

        :param Rb:
        :type Rb: float
        :param Zb:
        :type Zb: float
        :return: float
        """

        r = sqrt(Zb**2 + (self.rtang - Rb)**2)
        if r > self.beamWidth / 2.:
            return 0.
        else:
            expTermUp = -1. * (r**2 / self.gaussR**2)
            expTermDown = -1. * ((self.beamWidth / 2.)**2) / (self.gaussR**2)
            result = exp(expTermUp)/(pi * self.gaussR**2 * (1. - exp(expTermDown)))
            return result

    def hofrpz(self, Rb, Zb, rho_val, mfp, p, smallFlag):
        """
        The h(rho, Rb, Zb) function.

        :rtype: float
        :param p: Whether this is for the positive or negative half of the integration
        :type p: bool
        :param smallFlag: Flag to indicate that we are at small rho and need to address the singularity
        :type smallFlag: bool
        :return: The numeric value of this function
        :param mfp: The mean free path interpolator for this plasma energy group
        :type mfp: UnivariateSpline
        :param Rb: The beam-R coordinate
        :type float
        :param Zb: The beam-Z coordinate
        :type Zb: float
        :param rho_val: The rho value (0. < rho < 1.)
        :type float

        :return:
        """


        # TODO: This next lien is as in the manual, but NBEAMS was a subtraction
        Rin = self.R0 + self.shift(self.rho[-1]) + sqrt(self.a**2 - Zb**2 / self.kappa(self.a)**2)
        """ The R that the beam entered the plasma at (the plasma edge basically)"""

        lambda_mfp = mfp  # type: UnivariateSpline
        gamma = 1
        r = rho_val * self.a

        if p:
            # This probably should be updated for negative triangularity
            if (r)**2 - Zb**2/(self.kappa(rho_val)**2) >= 0:
                Rplus = self.get_Rpm(Zb, rho_val, True)
                """The largest R of this flux surface"""
                if Rplus:
                    # Is the beam R greater than the minimum R from the tangency radius on the inboard side
                    if Rplus > (self.rtang - sqrt((self.beamWidth / 2.) ** 2 - Zb**2)):
                        D0plus = self.D(Rplus, Rin, Zb, mfp)
                        if Rb >= Rin:
                            D1plus = self.D(Rb, Rplus, Zb, mfp)
                        else:
                            D1plus = 0.
                            gamma = 0.
                        # Updated to use r2vol instead of full plasma volume since that seems to make more sense.
                        # Doing self.rho[-2] beacuse when the separatrix is manually defined, the interpolator seems
                        # to mess up that very last value, so we push inward slightly.
                        #part1plus = ((2 * r * self.r2vol(self.a)) / self.dVdr(r)) * (1 / lambda_mfp(rho_val)) * (Rplus / (sqrt(Rplus ** 2 - self.rtang ** 2)))
                        part1plus = ((2 * r * self.r2vol(self.a * self.rho[-2])) / self.dVdr(r)) * (1 / lambda_mfp(rho_val)) * (Rplus / (sqrt(Rplus ** 2 - Rb ** 2)))
                        if smallFlag:
                            part2plus = 1.
                        else:
                            part2plus = np.nan_to_num((1 + (self.kappa_prime(rho_val) / (r)) * ((Zb ** 2) / self.kappa(rho_val) ** 3)) / (
                                sqrt(r ** 2 - (Zb ** 2) / (self.kappa(rho_val) ** 2))) + (self.shift_prime(rho_val) / (r)))
                        part3plus = exp(-1. * D0plus) + gamma * exp(-1. * (D0plus + 2.0 * D1plus))

                        hplus = part1plus * part2plus * part3plus
                        if self.verbose: print "Posiitve: Part 1: %s" % part1plus
                        if self.verbose: print "Posiitve: Part 2: %s" % part2plus
                        if self.verbose: print "Posiitve: Part 3: %s" % part3plus
                        if self.dVdr(r) <= 0.:
                            warnings.warn("Beams Module - Warning - dV/dr is negative at rho = %s" % rho_val)
                            hplus = 0.
                        if hplus < 0.:
                            raise ValueError("Beams Module - Error - hplus is negative at rho = %s" % rho_val)
                        if isinf(hplus):
                            raise ValueError("Beams Module - Erorr - hplus is infinite at rho = %s" % rho_val)
                        #print "Positive result: %s" % hplus

                        return hplus
                    else:
                        return 0
                else:
                    return 0
            else:
                return 0
        else:
            if (r)**2 - Zb**2/(self.kappa(rho_val)**2) >= 0:
                Rminus = self.get_Rpm(Zb, rho_val, False)
                if Rminus:
                    #if Rminus > (self.rtang - sqrt((self.beamWidth / 2.) ** 2 - Zb**2)):
                    if Rminus > (Rb - sqrt((self.beamWidth / 2.) ** 2 - Zb ** 2)):
                        D0minus = self.D(Rminus, Rin, Zb, mfp)
                        if Rb >= Rin:
                            D1minus = self.D(Rb, Rminus, Zb, mfp)
                        else:
                            D1minus = 0.
                            gamma = 0.
                        #part1minus = ((2 * (r) * self.r2vol(self.a)) / self.dVdr(r)) * (1 / lambda_mfp(rho_val)) * (Rminus / (sqrt(Rminus**2 - self.rtang**2)))
                        part1minus = ((2 * (r) * self.r2vol(self.a)) / self.dVdr(r)) * (1 / lambda_mfp(rho_val)) * (Rminus / (sqrt(Rminus ** 2 - Rb ** 2)))
                        if smallFlag:
                            part2minus = 1.
                        else:
                            part2minus = (1 + (self.kappa_prime(rho_val) / (r)) * ((Zb**2)/self.kappa(rho_val)**3))/(sqrt((r)**2 - (Zb**2)/(self.kappa(rho_val)**2))) + (self.shift_prime(rho_val)/(r))
                        part3minus = exp((-1. * D0minus)) + gamma * exp(-1. * (D0minus + 2.0 * D1minus))

                        hminus = part1minus + part2minus + part3minus
                        if self.dVdr(r) <= 0.:
                            warnings.warn("Beams Module - Warning - dV/dr is negative at rho = %s" % rho_val)
                            hminus = 0.
                        if hminus < 0.:
                            raise ValueError("Beams Module - Error - hminus is negative at rho = %s" % rho_val)
                        if isinf(hminus):
                            raise ValueError("Beams Module - Error - hminus is infinite at rho = %s" % rho_val)
                        if self.verbose: print "Negative: Part 1: %s" % part1minus
                        if self.verbose: print "Negative: Part 2: %s" % part2minus
                        if self.verbose: print "Negative: Part 3: %s" % part3minus
                        if self.verbose: print "Negative result: %s" % hminus
                        return hminus
                    else:
                        return 0
                else:
                    return 0
            else:
                return 0

    def get_Rpm(self, Zb, rho_val, p):
        """
        Get R_+- for a given (rho, Zb). This is the major radius of the outer (+) and inner (-) extents of the given
        flux surface a distance Zb above the midplane.

        :param Zb: The Zb coordinate
        :type Zb:  float
        :param rho_val: The rho value
        :type rho_val:  float
        :param p: Is this positive (True) or negative (False)
        :type p:  bool
        :return:
        """
        if p:
            if (rho_val*self.a)**2 - Zb**2/(self.kappa(rho_val)**2) <= 0:
                return False
            else:
                return self.R0 + self.shift(rho_val) + sqrt((rho_val*self.a)**2 - (Zb**2)/(self.kappa(rho_val)**2))
        else:
            if (rho_val*self.a)**2 - (Zb**2)/(self.kappa(rho_val)**2) <= 0:
                return False
            else:
                return self.R0 + self.shift(rho_val) - sqrt((rho_val*self.a)**2 - (Zb**2)/(self.kappa(rho_val)**2))

    def D(self, R1, R2, Zb, mfp):
        """
        This is the D function

        :param R1: The lower bound of the integration
        :type R1: float
        :param R2: The upper bound of the integration
        :type R2: float
        :param Zb: The Zb value that this integration is being performed at
        :type Zb: float
        :param mfp: The mean free path interpolator function passed to the Lambda function
        :type mfp: UnivariateSpline
        :return: The numeric value for this calculation
        :rtype: float
        """
        DfunLambda = lambda x, mfp=mfp: self.Dfun(x, Zb, mfp)

        if self.quad == 'quad':
            result = quad(DfunLambda, R1, R2, epsrel=.05)
            return result[0]

        elif self.quad == 'fixed_quad':
            self.prof_D['count'] += 1
            now = time()
            try:
                result = fixed_quad(DfunLambda, R1, R2, n=2)[0]
            except:
                print "Something went wrong"
            if result < 0.:
                raise ValueError("Attenuation D is negative")
            self.prof_D['time'].append(time() - now)
            return result

        elif self.quad == 'trap':
            try:
                return cumtrapz([DfunLambda(a) for a in np.linspace(R1, R2, 10)], np.linspace(R1, R2, 10))[-1]
            except:
                print "Something went wrong"

    def Dfun(self, x, Zb, mfp):
        """
        The function for use in the integration for calculating D. This takes Rb as a float or array depending on
        the integration scheme.

        :param x: The Rb value for the D integration at this point
        :type x: float
        :param Zb: The Zb value for the D integration at this point
        :type Zb: float
        :param mfp: The mean free path interpolator
        :type mfp: UnivariateSpline
        :return: The result of this function
        """
        if type(x) == np.ndarray:
            return[(a / (sqrt(a**2 - self.rtang**2))) * (1 / self.lam_special(a, Zb, mfp)) for a in x]
        else:
            return (x / (sqrt(x**2 - self.rtang**2))) * (1 / self.lam_special(x, Zb, mfp))

    def lam_special(self, R, Zb, mfp):
        """
        Attempts to find the special lambda for this R, Zb coordinate for this beam energy

        :return: Returns the mean free path in m
        :param mfp: The mean free path interpolator
        :type mfp: UnivariateSpline
        :param R: The major radius coordinate of the beam
        :type R:  float
        :param Zb: The Z coordinate of the beam
        :type Zb:  float
        :return:
        """
        now = time()
        fluxrho_l = lambda x: self.fluxrho(x, R, Zb)
        try:
            result = root_scalar(fluxrho_l, bracket=[0., self.sep_val]).root
        except Exception as e:
            print "Problem with root calculation: %s" % str(e)
            result = .9

        self.prof_root['count'] += 1
        self.prof_root['time'].append(time() - now)
        now = time()
        special_mfp = mfp(result)

        self.prof_mfp['count'] += 1
        self.prof_mfp['time'].append(time() - now)
        return special_mfp

    def fluxrho(self, x, R, Zb):
        """
        The F(x) = 0 function for calculating the flux surface that corresponds to the major radius in Eq (5)

        :param x: The rho (in [0., 1.]) value
        :type x:  float
        :param R: The R value from the integration
        :type R:  float
        :param Zb: The Zb value from the integration
        :type Zb:  float
        :return:  The rho value of the flux surface
        """
        return (self.a * x)**2 - (R - (self.R0 + self.shift(x)))**2 - Zb**2 / (self.kappa(x)**2)

    def plot_mfp(self):
        mfp_1 = get_mfp(self.ne, self.beamE, self.Te, self.beamA, self.Zeff, self.rho)
        mfp_2 = get_mfp(self.ne, self.beamE / 2., self.Te, self.beamA, self.Zeff, self.rho)
        mfp_3 = get_mfp(self.ne, self.beamE / 3., self.Te, self.beamA, self.Zeff, self.rho)

        mfp_l_1 = lambda x: mfp_1(x)
        mfp_l_2 = lambda x: mfp_2(x)
        mfp_l_3 = lambda x: mfp_3(x)
        fig = plt.figure()
        fig1 = fig.add_subplot(111)
        fig1.set_xlabel(r'$/rho$')
        fig1.set_ylabel(r'$\lambda [m]$')
        fig1.scatter(self.rho, [mfp_l_1(x) for x in self.rho], color='red')
        fig1.scatter(self.rho, [mfp_l_2(x) for x in self.rho], color='blue')
        fig1.scatter(self.rho, [mfp_l_3(x) for x in self.rho], color='green')
        plt.show()
        return fig1

    def plot_xs(self):
        sig_int_1 = sigeff(self.ne, self.beamE, self.Te, self.beamA, self.Zeff, self.rho)
        sig_int_2 = sigeff(self.ne, self.beamE / 2., self.Te, self.beamA, self.Zeff, self.rho)
        sig_int_3 = sigeff(self.ne, self.beamE / 3., self.Te, self.beamA, self.Zeff, self.rho)
        fig = plt.figure()
        fig1 = fig.add_subplot(111)
        fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
        fig1.set_xlabel(r'$/rho$')
        fig1.set_ylabel(r'$\sigma_{eff} [cm^2]$')
        fig1.set_ylim(1E-15, 1E-16)
        fig1.scatter(self.rho, sig_int_1(self.rho), color="red")
        fig1.scatter(self.rho, sig_int_2(self.rho), color="blue")
        fig1.scatter(self.rho, sig_int_3(self.rho), color="green")
        fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
        plt.show()
        return fig1

    def plot_HofRho(self):

        fig = plt.figure()
        fig1 = fig.add_subplot(111)
        #fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
        fig1.set_xlabel(r'$/rho$')
        fig1.set_ylabel(r'$H(\rho)$')
        #fig1.set_ylim(np.min(y, 0), np.max(y, 0))
        fig1.scatter(self.rho, self.Hofrho.D1(self.rho), color='red')
        fig1.scatter(self.rho, self.Hofrho.D2(self.rho), color='blue')
        fig1.scatter(self.rho, self.Hofrho.D3(self.rho), color='green')
        plt.show()
        return fig1
        #fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))

    def plot_Hofr(self):
        fig = plt.figure()
        fig1 = fig.add_subplot(111)
        # fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
        fig1.set_xlabel(r'$/rho$')
        fig1.set_ylabel(r'$H(r)$')
        fig1.scatter(self.rho, self.Hofr.D1(self.rho / self.a), color='red')
        fig1.scatter(self.rho, self.Hofr.D2(self.rho / self.a), color='blue')
        fig1.scatter(self.rho, self.Hofr.D3(self.rho / self.a), color='green')
        plt.show()
        return fig1
        # fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))

    def plot_power(self):

        fig = plt.figure()
        fig1 = fig.add_subplot(111)
        #fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
        fig1.set_xlabel(r'$/rho$')
        fig1.set_ylabel(r'$\frac{dP}{dV}_{nb}(\rho) [W/{m^3}]$')
        #fig1.set_ylim(np.min(y, 0), np.max(y, 0))
        fig1.scatter(self.rho, self.dPdV.D1(self.rho), color='red')
        fig1.scatter(self.rho, self.dPdV.D2(self.rho), color='blue')
        fig1.scatter(self.rho, self.dPdV.D3(self.rho), color='green')
        fig1.scatter(self.rho, self.dPdV.total(self.rho), color='black')
        #fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
        plt.show()
        return fig1

    def plot_hofRho(self, rho, energy):
        r = np.linspace(1, 1.3, 50)
        z = np.linspace(0., .1, 50)
        R, Z = np.meshgrid(r, z)
        f = lambda x, y: self.hofrpzJ(x, y, rho, energy,  True) + self.hofrpzJ(x, y, rho, energy, False)
        values = np.zeros((50, 50))
        for i in range(50):
            for j in range(50):
                values[i][j] = f(z[i], r[j])

        fig = plt.figure()
        ax = plt.axes(projection='3d')
        ax.scatter(R, Z, values, cmap='viridis', linewidth=0.5)
        plt.show()
        return fig, ax

    def plot_hofrho_unnorm(self):

        fig = plt.figure()
        fig1 = fig.add_subplot(111)
        #fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
        fig1.set_xlabel(r'$/rho$')
        fig1.set_ylabel(r'$P_{nb}(\rho)$')
        #fig1.set_ylim(np.min(y, 0), np.max(y, 0))
        fig1.scatter(self.rho, self.Hofrho_unnorm.D1(self.rho), color='red')
        fig1.scatter(self.rho, self.Hofrho_unnorm.D2(self.rho), color='blue')
        fig1.scatter(self.rho, self.Hofrho_unnorm.D3(self.rho), color='green')
        plt.show()
        return fig1
        #fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))

    def plot_J(self):
        r = np.linspace(1, 1.3, 50)
        z = np.linspace(0., .1, 50)
        R, Z = np.meshgrid(r, z)
        f = lambda x, y: self.J(x, y)
        values = np.zeros((50, 50))
        for i in range(50):
            for j in range(50):
                values[i][j] = f(r[i], z[j])

        fig = plt.figure()
        ax = plt.axes(projection='3d')
        ax.scatter(R, Z, values, cmap='viridis', linewidth=0.5)
        plt.show()
        return fig, ax

    def plot_part_src_kept(self):
        fig = plt.figure()
        fig1 = fig.add_subplot(111)
        #fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
        fig1.set_xlabel(r'$r$')
        fig1.set_ylabel(r'$S_{nb}(r) [#/m^3 s]$')
        #fig1.set_ylim(np.min(y, 0), np.max(y, 0))
        fig1.scatter(self.rho * self.a, self.part_src_D1.kept, color='red')
        fig1.scatter(self.rho * self.a, self.part_src_D2.kept, color='blue')
        fig1.scatter(self.rho * self.a, self.part_src_D3.kept, color='green')
        fig1.scatter(self.rho * self.a, self.part_src_total.kept, color='green')
        plt.show()
        return fig1
        #fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))

    def plot_energy_src_kep(self):
        fig = plt.figure()
        fig1 = fig.add_subplot(111)
        # fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
        fig1.set_xlabel(r'$/rho$')
        fig1.set_ylabel(r'$Q_{nb}(r) [W/m^3]$')
        # fig1.set_ylim(np.min(y, 0), np.max(y, 0))
        fig1.scatter(self.rho * self.a, self.en_src_D1.kept, color='red')
        fig1.scatter(self.rho * self.a, self.en_src_D2.kept, color='blue')
        fig1.scatter(self.rho * self.a, self.en_src_D3.kept, color='green')
        fig1.scatter(self.rho * self.a, self.en_src_dens, color='black')
        plt.show()
        return fig1
        # fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))

    def plot_mom_src_kept(self):
        fig = plt.figure()
        fig1 = fig.add_subplot(111)
        #fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
        fig1.set_xlabel(r'$r')
        fig1.set_ylabel(r'$M_{nb}(r)$')
        #fig1.set_ylim(np.min(y, 0), np.max(y, 0))
        fig1.scatter(self.rho * self.a, self.mom_src_D1.kept, color='red')
        fig1.scatter(self.rho * self.a, self.mom_src_D2.kept, color='blue')
        fig1.scatter(self.rho * self.a, self.mom_src_D3.kept, color='green')
        fig1.scatter(self.rho * self.a, self.mom_src_tot_kept, color='black')
        plt.show()
        return fig1
        #fig1.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))

    def plot_iol(self):
        fig = plt.figure()
        fig1 = fig.add_subplot(311)
        fig1.set_xlabel(r'$/rho$')
        fig1.set_ylabel(r'$F_{iol}(\rho)$')
        fig1.scatter(self.rho, self.F_orb_D1, color='red')
        fig1.scatter(self.rho, self.F_orb_D2, color='blue')
        fig1.scatter(self.rho, self.F_orb_D3, color='green')

        fig2 = fig.add_subplot(312)
        fig2.set_xlabel(r'$/rho$')
        fig2.set_ylabel(r'$E_{iol}(\rho)$')
        fig2.scatter(self.rho, self.E_orb_D1, color='red')
        fig2.scatter(self.rho, self.E_orb_D2, color='blue')
        fig2.scatter(self.rho, self.E_orb_D3, color='green')

        fig3 = fig.add_subplot(313)
        fig3.set_xlabel(r'$/rho$')
        fig3.set_ylabel(r'$M_{iol}(\rho)$')
        fig3.scatter(self.rho, self.M_orb_D1, color='red')
        fig3.scatter(self.rho, self.M_orb_D2, color='blue')
        fig3.scatter(self.rho, self.M_orb_D3, color='green')

        return fig




#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy import constants
from scipy.interpolate import UnivariateSpline

e = constants.elementary_charge


def calc_chi_jet(T, L, a, q, B_T, m_i, rho):
    def calc_bohm(T, L, a, q, B_T, m_i, rho):
        """"""
        cs = np.sqrt(T.i.J / m_i)  # sound speed
        rho_s = cs * m_i / (e * B_T)  # gyroradius

        Te_interp = UnivariateSpline(rho[:, 0], T.e.J[:, 0], k=3, s=0)
        delta_Te = (Te_interp(0.8) - Te_interp(1.0)) / Te_interp(1.0)
        chi_bohm = rho_s * cs * q ** 2 * a * L.p.e * delta_Te
        return chi_bohm

    def calc_gyro_bohm(T, L, B_T, m_i):
        cs = np.sqrt(T.i.J / m_i)  # sound speed
        rho_s = cs * m_i / (e * B_T)  # gyroradius
        chi_gyro_bohm = rho_s ** 2 * cs * L.T.e
        return chi_gyro_bohm

    chi_bohm = calc_bohm(T, L, a, q, B_T, m_i, rho)
    chi_gyro_bohm = calc_gyro_bohm(T, L, B_T, m_i)
    chi_i_jet = 1.6E-4 * chi_bohm + 1.75E-2 * chi_gyro_bohm
    chi_e_jet = 8E-5 * chi_bohm + 3.5E-2 * chi_gyro_bohm
    return chi_i_jet, chi_e_jet



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Sat Aug  5 16:10:25 2017

@author: max
"""
import os
import sys
import re
import matplotlib.pyplot as plt
import numpy as np
from math import pi
import ConfigParser

from shapely.geometry import LineString


class ReadInfile:
    """Reads main GT3 input file.

    Methods:
        read_vars
        read_exp
        showparams

    Attributes:
        a                (float)    tokamak minor radius (m)
        BT0          (float)    toroidal field strength at mag. axis (T)
        R0_a             (float)    tokamak major radius (m)
        Z0               (float)    vertical height of the magnetic axis (m)
        xpt_R            (float)
        xpt_Z            (float)
        xpt
        thetapts_approx  (int)
        thetapts
        rmeshnum_p       (int)
        rpts             (int)
        ni0              (float)
        ni9              (float)
        ni_sep           (float)
        nu_ni            (float)
        ne0              (float)
        ne9              (float)
        ne_sep           (float)
        nu_ne            (float)
        Ti0              (float)
        Ti9              (float)
        Ti_sep           (float)
        nu_Ti            (float)
        Te0              (float)
        Te9              (float)
        Te_sep           (float)
        nu_Te            (float)
        j0               (float)
        wallfile         (str)
    """

    def __init__(self, infile):
        """
        Initializes the read_infile class with the filename for the to_gt3 file.
        :param infile:
        :type infile: str
        """
        sys.dont_write_bytecode = True
        self.read_vars(infile)

    def _profile_loader(self, parser, section, name):
        """ A helper function for loading profiles.

        :param parser: The ConfigParser
        :type parser: ConfigParser.RawConfigParser
        :param section: The section name
        :type section: str
        :param name: The variable name
        :type name: str
        :return:
        """

        try:
            filename = parser.get(section, name)
            filepath = os.path.join(os.getcwd(), filename)
            return np.genfromtxt(filepath, comments='#')
        except ConfigParser.NoOptionError:
            print "%s not found" % name
            return
        except IOError:
            print "%s not found" % name
            return

    def read_vars(self, infile):
        """
        Reads variables given infile filename
        :param infile:
        :type infile: str
        """

        config = ConfigParser.RawConfigParser()
        config.read(infile)

        # Grid construction parameters

        self.rhopts = config.getint('Mesh', 'rhopts')
        self.edge_rho = config.getfloat('Mesh', 'edge_rho')
        self.rhopts_edge = config.getint('Mesh', 'rhopts_edge')
        self.rhopts_core = config.getint('Mesh', 'rhopts_core')
        self.thetapts_approx = config.getint('Mesh', 'thetapts_approx')
        try:
            self.Er_scale = config.getfloat('Mesh', 'Er_scale')
        except ConfigParser.NoOptionError:
            self.Er_scale = 1.
        try:
            self.psi_scale = config.getfloat('Mesh', 'psi_scale')
        except ConfigParser.NoOptionError:
            self.psi_scale = 1.
        self.sollines_psi_max = config.getfloat('Mesh', 'sollines_psi_max')
        self.num_sollines = config.getint('Mesh', 'num_sollines')
        self.xi_ib_pts = config.getint('Mesh', 'xi_ib_pts')
        self.xi_ob_pts = config.getint('Mesh', 'xi_ob_pts')
        self.numcos = config.getint('Mesh', 'numcos')


        # Plasma
        self.BT0 = config.getfloat('Plasma', 'Bt0')
        self.pfr_ni_val = config.getfloat('Plasma', 'pfr_ni_val')
        self.pfr_ne_val = config.getfloat('Plasma', 'pfr_ne_val')
        self.pfr_Ti_val = config.getfloat('Plasma', 'pfr_Ti_val')
        self.pfr_Te_val = config.getfloat('Plasma', 'pfr_Te_val')
        self.R_loss = config.getfloat('Plasma', 'R_loss')
        try:
            self.sep_val = config.getfloat('Plasma', 'sep_val')
        except ConfigParser.NoOptionError:
            self.sep_val = 1.0


        # Neutrals Output File
        self.neutfile_loc = config.get('1DProfiles', 'neutfile_loc')

        # NBeams multi-beam location
        self.beams_json = config.get('1DProfiles', 'beams_json')
        self.beams_out_json = config.get('1DProfiles', 'beams_out_json')

        self.er_data = self._profile_loader(config, '1DProfiles', 'er_file')
        self.jr_data = self._profile_loader(config, '1DProfiles', 'jr_file')
        self.ne_data = self._profile_loader(config, '1DProfiles', 'ne_file')
        self.nD_data = self._profile_loader(config, '1DProfiles', 'nD_file')
        self.nT_data = self._profile_loader(config, '1DProfiles', 'nT_file')
        self.nW_data = self._profile_loader(config, '1DProfiles', 'nW_file')
        self.nBe_data = self._profile_loader(config, '1DProfiles', 'nBe_file')
        self.na_data = self._profile_loader(config, '1DProfiles', 'na_file')
        self.nC_data = self._profile_loader(config, '1DProfiles', 'nC_file')
        self.Te_data = self._profile_loader(config, '1DProfiles', 'Te_file')
        self.Ti_data = self._profile_loader(config, '1DProfiles', 'Ti_file')
        self.TC_data = self._profile_loader(config, '1DProfiles', 'TC_file')
        self.frac_C_data = self._profile_loader(config, '1DProfiles', 'frac_C_file')
        self.vpolC_data = self._profile_loader(config, '1DProfiles', 'vpolC_file')
        self.vtorC_data = self._profile_loader(config, '1DProfiles', 'vtorC_file')
        self.vpolD_data = self._profile_loader(config, '1DProfiles', 'vpolD_file')
        self.vtorD_data = self._profile_loader(config, '1DProfiles', 'vtorD_file')

        self.psirz_exp = self._profile_loader(config, '2DProfiles', 'psirz_file')
        self.wall_exp = self._profile_loader(config, 'Wall', 'wall_file')

        self.wall_line = LineString(self.wall_exp)


    def showparams(self):
        """
        Show the parameters of this shot
        """
        print '**PARAMETERS FOR SHOT \'{}\'.'.format(self.shotlabel)
        for key in vars(self).iteritems():
            if key[0][0] != '_' and key[0] != 'line' and key[0] != 'infile' and key[0] != 'variable' and key[
                0] != 'value':
                print ('{} = {}'.format(key[0], key[1]))
        print '**END OF PARAMETERS**'

    def _plot_base(self, xLabel=r'$\rho$', yLabel="Value", title="Title", edge=False):

        plot = plt.figure()
        fig = plot.add_subplot(111)
        fig.set_xlabel(xLabel, fontsize=20)
        fig.set_ylabel(yLabel, fontsize=20)
        fig.set_title(title)
        if edge:
            fig.set_xlim(0.85, 1.0)
        plt.show()
        return fig

    def plot_raw_Ti(self, edge=False):
        fig = self._plot_base(edge=edge)
        fig.scatter(self.Ti_data.T[0], self.Ti_data.T[1], color="black", s=8)
        return fig

    def plot_raw_Te(self, edge=False):
        fig = self._plot_base(edge=edge)
        fig.scatter(self.Te_data.T[0], self.Te_data.T[1], color="black", s=8)
        return fig

    def plot_raw_Er(self, edge=False):
        fig = self._plot_base(edge=edge)
        fig.scatter(self.er_data.T[0], self.er_data.T[1], color="black", s=8)
        return fig

    def plot_raw_ni(self, edge=False):
        fig = self._plot_base(edge=edge)
        fig.scatter(self.nD_data.T[0], self.nD_data.T[1], color="black", s=8)
        return fig

    def plot_raw_ne(self, edge=False):
        fig = self._plot_base(edge=edge)
        fig.scatter(self.ne_data.T[0], self.ne_data.T[1], color="black", s=8)
        return fig

    def plot_raw_vpol_C(self, edge=False):
        fig = self._plot_base(edge=edge)
        fig.scatter(self.vpolC_data.T[0], self.vpolC_data.T[1], color="black", s=8)
        return fig

    def plot_raw_vpol_D(self, edge=False):
        fig = self._plot_base(edge=edge)
        fig.scatter(self.vpolD_data.T[0], self.vpolD_data.T[1], color="black", s=8)
        return fig

    def plot_raw_vtor_C(self, edge=False):
        fig = self._plot_base(edge=edge)
        fig.scatter(self.vtorC_data.T[0], self.vtorC_data.T[1], color="black", s=8)
        return fig

    def plot_raw_vtor_D(self, edge=False):
        fig = self._plot_base(edge=edge)
        fig.scatter(self.vtorD_data.T[0], self.vtorD_data.T[1], color="black", s=8)
        return fig



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from Test.DebugCore import debugCore
from Test.DebugInp import debugInp
import numpy as np
from scipy.interpolate import UnivariateSpline
import matplotlib.pyplot as plt
from collections import namedtuple
from scipy.constants import physical_constants
import os, sys
from Functions.Beam import Beam
import json
import yaml
from pathos.multiprocessing import ProcessPool as Pool
from pathos.multiprocessing import cpu_count
import GT3.constants as constants

m_d = constants.deuteron_mass
z_d = 1
MARKERSIZE = constants.MARKERSIZE


class BeamDeposition:
    beam_result = None  # type: list[Beam]
    beams = None  # type: list[Beam]

    BeamResult = namedtuple('BeamResult', 'rtang beamWidth gaussR beamE beamA beamP rho Hofrho_D1 Hofrho_D2 '
                                          'Hofrho_D3 angle_D1 angle_D2 angle_D3 shine')
    BeamConfiguration = namedtuple('BeamConfiguration', 'name rho r2vol dVdr dVdrho shaf_shift kappa_vals a R0'
                                                        ' Te Ti ne Z_eff rtang beamWidth gaussR beamE beamA '
                                                        'beamP iol coCurr verbose timing iolFlag '
                                                        'beamResult pwrFracOverride')

    BeamSources = namedtuple('BeamSources', 'Snbi Qnbi Mnbi')

    def __init__(self, inp, core, iol=None, reRun=False, pwrFracOverride=None):
        sys.dont_write_bytecode = True
        # If deposition profile is provided as an input, use that.
        # Note, this is designed to read in the dP/dr(r) quantity, not dP/dV(r).
        # When integrated over r, this should equal P_beam, or at least the amount
        # that is deposited in the plasma.

        self.beams_space = np.linspace(0., core.sep_val, 50)
        """The rho space used by the Beams module on [0., 1.] or [0., sep_val]"""

        self.pwrFracOverride = pwrFracOverride
        """Allows us to override the beam power fraction for sensitivity studies"""

        if not reRun:
            # nbi_vals = calc_nbi_vals(inp, core)
            # Look for NBI output file
            try:
                self.load_beams(inp, core, iol)
            except BaseException as e:
                print "Failed to load beams data. Running beams module..."
                self.call_beams(inp, core, iol)
        else:
            self.call_beams(inp, core, iol)

        try:
            # Combine into one beam.

            self.combined_beam_src_total = self.BeamSources(np.array([a.part_src.total for a in self.beam_result]).sum(axis=0),
                                          np.array([a.en_src.total for a in self.beam_result]).sum(axis=0),
                                          np.array([a.mom_src.total for a in self.beam_result]).sum(axis=0))

            self.combined_beam_src_lost = self.BeamSources(np.array([a.part_src.lost for a in self.beam_result]).sum(axis=0),
                                          np.array([a.en_src.lost for a in self.beam_result]).sum(axis=0),
                                          np.array([a.mom_src.lost for a in self.beam_result]).sum(axis=0))

            self.combined_beam_src_kept = self.BeamSources(np.array([a.part_src.kept for a in self.beam_result]).sum(axis=0),
                                          np.array([a.en_src.kept for a in self.beam_result]).sum(axis=0),
                                          np.array([a.mom_src.kept for a in self.beam_result]).sum(axis=0))

            self.combined_beam_src_dens_total = self.BeamSources(np.array([a.part_src_dens.total for a in self.beam_result]).sum(axis=0),
                                          np.array([a.en_src_dens.total for a in self.beam_result]).sum(axis=0),
                                          np.array([a.mom_src_dens.total for a in self.beam_result]).sum(axis=0))

            self.combined_beam_src_dens_lost = self.BeamSources(np.array([a.part_src_dens.lost for a in self.beam_result]).sum(axis=0),
                                          np.array([a.en_src_dens.lost for a in self.beam_result]).sum(axis=0),
                                          np.array([a.mom_src_dens.lost for a in self.beam_result]).sum(axis=0))

            self.combined_beam_src_dens_kept = self.BeamSources(np.array([a.part_src_dens.kept for a in self.beam_result]).sum(axis=0),
                                          np.array([a.en_src_dens.kept for a in self.beam_result]).sum(axis=0),
                                          np.array([a.mom_src_dens.kept for a in self.beam_result]).sum(axis=0))
        except Exception as e:
            raise Exception("NBI module failed to create main beam profiles with error: %s" % str(e))

    def load_beams(self, inp, core, iol):
        try:
            self.beam_result = []
            with open(os.path.join(os.getcwd(), inp.beams_out_json), "r") as f:
                l = yaml.safe_load(f)
                for a in l.keys():
                    beam = l[a]
                    self.beam_result.append(Beam(self.BeamConfiguration('name',
                                                                        self.beams_space,
                                                                        core.r2vol,
                                                                        core.dVdr,
                                                                        core.dVdrho,
                                                                        core.shaf_shift,
                                                                        core.kappa_vals,
                                                                        core.a,
                                                                        core.R0_a,
                                                                        UnivariateSpline(core.rho[:, 0],
                                                                                         core.T_fsa.e.kev)(
                                                                            self.beams_space),
                                                                        UnivariateSpline(core.rho[:, 0],
                                                                                         core.T_fsa.i.kev)(
                                                                            self.beams_space),
                                                                        UnivariateSpline(core.rho[:, 0], core.n_fsa.e)(
                                                                            self.beams_space),
                                                                        UnivariateSpline(core.rho[:, 0],
                                                                                         core.z_eff_fsa)(
                                                                            self.beams_space),
                                                                        beam[0],
                                                                        beam[1],
                                                                        beam[2],
                                                                        beam[3],
                                                                        beam[4],
                                                                        beam[5],
                                                                        iol,
                                                                        coCurr=True,
                                                                        verbose=False,
                                                                        timing=False,
                                                                        iolFlag=True,
                                                                        beamResult=beam,
                                                                        pwrFracOverride=self.pwrFracOverride)))

        except IOError as e:
            raise Exception("Beams output file IO error: %s" % str(e))
        except Exception as e:
            raise Exception("Failed to load beams output file: %s" % str(e))

    def call_beams(self, inp, core, iol):
        # Call signature: def calcHofRho(rho, r2vol, dVdr, rtang, shaf_shift, kappa_vals, beamHeight, beamWidth, gaussR, Te, TC, R0, ne, beamE, Zeff, a)

        try:
            try:
                f = open(os.path.join(os.getcwd(), inp.beams_json), "r")
            except IOError as error:
                raise IOError("Could not open Beams JSON. Error: %s" % str(error))
            newdata = yaml.safe_load(f)
            f.close()
            if type(newdata) is not dict:
                raise TypeError("Beams JSON data not loaded as dictionary")
            else:
                beam_config_list = []
                for n in newdata.keys():
                    d = newdata[n]
                    if float(d['beamP']) <= 0. or float(d['beamE']) < .01:
                        continue
                    try:
                        beam_config = self.BeamConfiguration(n,
                                                             self.beams_space,
                                                             core.r2vol,
                                                             core.dVdr,
                                                             core.dVdrho,
                                                             core.shaf_shift,
                                                             core.kappa_vals,
                                                             core.a,
                                                             core.R0_g,
                                                             UnivariateSpline(core.rho[:, 0], core.T_fsa.e.kev)(
                                                                 self.beams_space),
                                                             UnivariateSpline(core.rho[:, 0], core.T_fsa.i.kev)(
                                                                 self.beams_space),
                                                             UnivariateSpline(core.rho[:, 0], core.n_fsa.e)(
                                                                 self.beams_space),
                                                             UnivariateSpline(core.rho[:, 0], core.z_eff_fsa)(
                                                                 self.beams_space),
                                                             d['rtang'],
                                                             d['beamWidth'],
                                                             d['gaussR'],
                                                             d['beamE'],
                                                             d['beamA'],
                                                             d['beamP'],
                                                             iol,
                                                             coCurr=True,
                                                             verbose=False,
                                                             timing=False,
                                                             iolFlag=True,
                                                             beamResult=None,
                                                             pwrFracOverride=self.pwrFracOverride,
                                                             )
                        beam_config_list.append(beam_config)
                    except Exception as e:
                        print "BeamConfiguration failed for beam %s with error: %s " % (n, str(e))
                pool = Pool(cpu_count() - 1)

                try:
                    try:
                        #Beam1 = Beam(beam_config_list[0])
                        #Beam2 = Beam(beam_config_list[1])
                        self.beam_result = pool.map(Beam, beam_config_list)
                    except:
                        raise Exception("Pathos failed to run beam")

                    """A list of the raw results returned by the multi-processing Pools module"""
                    if self.beam_result is None or len(self.beam_result) == 0:
                        raise Exception("Beams module failed to run beam")
                except Exception:
                    raise Exception("Beams module failed to run beam")

                # Write the beams output file
                if len(self.beam_result) > 0:
                    out_dict = {}

                    with open(inp.beams_out_json, "w") as f:
                        for b in self.beam_result:
                            out_dict[b.name] = self.BeamResult(b.rtang,
                                                               b.beamWidth,
                                                               b.gaussR,
                                                               b.beamE / 1.E3,
                                                               b.beamA,
                                                               b.beamP,
                                                               b.rho.tolist(),
                                                               b.Hofrho.D1(b.rho).tolist(),
                                                               b.Hofrho.D2(b.rho).tolist(),
                                                               b.Hofrho.D3(b.rho).tolist(),
                                                               b.angle_1(b.rho).tolist(),
                                                               b.angle_2(b.rho).tolist(),
                                                               b.angle_3(b.rho).tolist(),
                                                               b.shine)
                        json.dump(out_dict, f, indent=4)

        except:
            print "No Beams JSON file found. Using single-beam data"

    def _plot_single_val(self, title="Title", xLabel=r'$\rho$', yLabel="Y-Axis"):

        fig = plt.figure()
        fig1 = fig.add_subplot(111)
        fig1.set_xlabel(xLabel)
        fig1.set_ylabel(yLabel)
        fig1.set_title(title)
        return fig1

    def plot_S_nbi(self):
        fig = self._plot_single_val(title="NBI Particle Source", yLabel=r'$S_{nbi}[\#/{m s}]$')
        fig.scatter(self.beams_space, self.combined_beam_src_kept.Snbi, color='red', s=MARKERSIZE)
        plt.show()
        return fig

    def plot_Q_nbi(self):
        fig = self._plot_single_val(title="NBI Energy Source", yLabel=r'$Q_{nbi}[W/m]$')
        fig.scatter(self.beams_space, self.combined_beam_src_kept.Qnbi, color='red',s=MARKERSIZE)
        plt.show()
        return fig

    def plot_M_nbi(self):
        fig = self._plot_single_val(title="NBI Momentum Source", yLabel=r'$S_{nbi}[M/s}]$')
        fig.scatter(self.beams_space, self.combined_beam_src_kept.Qnbi, color='red', s=MARKERSIZE)
        plt.show()
        return fig

    def plot_S_dens_nbi(self):
        fig = self._plot_single_val(title="NBI Particle Source Density", yLabel=r'$S_{nbi}[\#/{m^3 s}]$')
        fig.scatter(self.beams_space, self.combined_beam_src_dens_kept.Snbi, color='red', s=MARKERSIZE)
        plt.show()
        return fig

    def plot_Q_dens_nbi(self):
        fig = self._plot_single_val(title="NBI Energy Source Density", yLabel=r'$Q_{nbi}[W/{m^3}]$')
        fig.scatter(self.beams_space, self.combined_beam_src_dens_kept.Qnbi, color='red', s=MARKERSIZE)
        plt.show()
        return fig

    def plot_M_dens_nbi(self):
        fig = self._plot_single_val(title="NBI Momentum Source Density", yLabel=r'$S_{nbi}[\frac{M}{s}]$')
        fig.scatter(self.beams_space, self.combined_beam_src_dens_kept.Qnbi, color='red', s=MARKERSIZE)
        plt.show()
        return fig



if __name__ == "__main__":
    import matplotlib.pyplot as plt

    inp = debugInp()
    core = debugCore()
    beamData = BeamDeposition(inp, core)
    print """Printing NBeams input parameters:
            Core
            {:<16}          {}
            {:<16}          {}
            {:<16}          {}
            {:<16}          {}
            {:<16}          {}
            {:<16}          {}
            """.format("R0_a", str(core.R0_a),
                       "a", str(core.a),
                       "kappa_axis", str(core.kappa.axis),
                       "kappa_sep", str(core.kappa.sep),
                       "pts.axis.mag[0]", str(core.pts.axis.mag[0]),
                       "shaf_shift", str(core.shaf_shift),
                       )

    print """
            inp           
            {:<16}          {}
            {:<16}          {}
            {:<16}          {}
            {:<16}          {}
            {:<16}          {}
            """.format("BT0", str(inp.BT0),
                       "ebeam", str(inp.ebeam),
                       "nbeams_loc", str(inp.nbeams_loc),
                       "pbeam", str(inp.pbeam),
                       "rtang", str(inp.rtang)
                       )

    print """
        D1 info
        {:<10}      {}
        {:<10}      {}
        {:<10}      {}
        """.format("m", str(beamData.beams.D1.m),
                   "rtan", str(beamData.beams.D1.rtan),
                   "zeta", str(beamData.beams.D1.zeta))
    print """
        D2 info
        {:<10}      {}
        {:<10}      {}
        {:<10}      {}
        """.format("m", str(beamData.beams.D2.m),
                   "rtan", str(beamData.beams.D2.rtan),
                   "zeta", str(beamData.beams.D2.zeta))
    print """
        D3 info
        {:<10}      {}
        {:<10}      {}
        {:<10}      {}
        """.format("m", str(beamData.beams.D3.m),
                   "rtan", str(beamData.beams.D3.rtan),
                   "zeta", str(beamData.beams.D3.zeta))

    print """Calculated nbeams values"""
    fig0 = plt.figure(figsize=(12, 8))
    fig0.suptitle(r"Deposition profiles")
    axi = 230
    for a in beamData.debug.keys():
        if type(beamData.debug[a]) is float:
            print "{:<10}      {}".format(str(a), str(beamData.debug[a]))
        elif type(beamData.debug[a]) is np.ndarray:
            if "hofr" in a:
                axi += 1
                ax = fig0.add_subplot(axi)
                ax.set_title(a, fontsize=16)
                ax.set_ylabel("dep", fontsize=16)
                ax.set_xlabel(r"rho", fontsize=16)
                ax.plot(np.linspace(0, 1, 51), beamData.debug[a])
    d1Ptot = UnivariateSpline(core.rho[:, 0], beamData.beams.D1.dPdr.v1D.W).integral(0., 1.)
    d2Ptot = UnivariateSpline(core.rho[:, 0], beamData.beams.D2.dPdr.v1D.W).integral(0., 1.)
    d3Ptot = UnivariateSpline(core.rho[:, 0], beamData.beams.D3.dPdr.v1D.W).integral(0., 1.)
    Ptot = d1Ptot + d2Ptot + d3Ptot
    print "{:<10}      {}".format("D1 Total power", str(d1Ptot))
    print "{:<10}      {}".format("D2 Total power", str(d2Ptot))
    print "{:<10}      {}".format("D3 Total power", str(d3Ptot))
    print "{:<10}      {}".format("Total power from dpdr graph integration", str(Ptot))

    fig = plt.figure(figsize=(12, 8))
    fig.suptitle(r'NBeams Debug Info')
    ax1 = fig.add_subplot(231)
    ax1.set_title(r'ion density', fontsize=16)
    ax1.set_ylabel(r'$n_i[1/{m^3}]$', fontsize=16)
    ax1.set_xlabel(r'rho', fontsize=16)
    ax1.plot(core.rho[:, 0], core.n.i[:, 0])

    ax2 = fig.add_subplot(232)
    ax2.set_title(r'electron density', fontsize=16)
    ax2.set_ylabel(r'$n_e[1/{m^3}]$', fontsize=16)
    ax2.set_xlabel(r'rho', fontsize=16)
    ax2.plot(core.rho[:, 0], core.n.e[:, 0])

    ax3 = fig.add_subplot(233)
    ax3.set_title(r'Ion temp (kev)', fontsize=16)
    ax3.set_ylabel(r'$T_i[kev}$', fontsize=16)
    ax3.set_xlabel(r'rho', fontsize=16)
    ax3.plot(core.rho[:, 0], core.T.i.kev[:, 0])

    ax4 = fig.add_subplot(234)
    ax4.set_title(r'Electron temp (kev)', fontsize=16)
    ax4.set_ylabel(r'$T_e[kev}$', fontsize=16)
    ax4.set_xlabel(r'rho', fontsize=16)
    ax4.plot(core.rho[:, 0], core.T.e.kev[:, 0])

    ax5 = fig.add_subplot(235)
    ax5.set_title(r'Electron temp (kev)', fontsize=16)
    ax5.set_ylabel(r'$T_e[keV]$', fontsize=16)
    ax5.set_xlabel(r'rho', fontsize=16)
    ax5.plot(core.rho[:, 0], core.T.e.kev[:, 0])

    fig2 = plt.figure(figsize=(12, 8))
    fig2.suptitle(r"""$D_1$ data""")

    ax21 = fig2.add_subplot(231)
    ax21.set_title(r'Power density profile', fontsize=16)
    ax21.set_ylabel(r'$\frac{dP}{dr}$', fontsize=16)
    ax21.set_xlabel(r'rho', fontsize=16)
    ax21.plot(core.rho[:, 0], beamData.beams.D1.dPdr.v1D.W)

    ax22 = fig2.add_subplot(232)
    ax22.set_title(r'Power density profile', fontsize=16)
    ax22.set_ylabel(r'$\frac{dP}{dV}$', fontsize=16)
    ax22.set_xlabel(r'rho', fontsize=16)
    ax22.plot(core.rho[:, 0], beamData.beams.D1.dPdV.v1D.W)

    fig3 = plt.figure(figsize=(12, 8))
    fig3.suptitle(r"""$D_2$ data""")

    ax31 = fig3.add_subplot(231)
    ax31.set_title(r'Power density profile', fontsize=16)
    ax31.set_ylabel(r'$\frac{dP}{dr}$', fontsize=16)
    ax31.set_xlabel(r'rho', fontsize=16)
    ax31.plot(core.rho[:, 0], beamData.beams.D2.dPdr.v1D.W)

    ax32 = fig3.add_subplot(232)
    ax32.set_title(r'Power density profile', fontsize=16)
    ax32.set_ylabel(r'$\frac{dP}{dV}$', fontsize=16)
    ax32.set_xlabel(r'rho', fontsize=16)
    ax32.plot(core.rho[:, 0], beamData.beams.D2.dPdV.v1D.W)

    fig4 = plt.figure(figsize=(12, 8))
    fig4.suptitle(r"""$D_3$ data""")

    ax41 = fig4.add_subplot(231)
    ax41.set_title(r'Power density profile', fontsize=16)
    ax41.set_ylabel(r'$\frac{dP}{dr}$', fontsize=16)
    ax41.set_xlabel(r'rho', fontsize=16)
    ax41.plot(core.rho[:, 0], beamData.beams.D3.dPdr.v1D.W)

    ax42 = fig4.add_subplot(232)
    ax42.set_title(r'Power density profile', fontsize=16)
    ax42.set_ylabel(r'$\frac{dP}{dV}$', fontsize=16)
    ax42.set_xlabel(r'rho', fontsize=16)
    ax42.plot(core.rho[:, 0], beamData.beams.D3.dPdV.v1D.W)

    ax43 = fig4.add_subplot(233)
    ax43.set_title(r'Power density profile', fontsize=16)
    ax43.set_ylabel(r'$\frac{dP}{dr}$', fontsize=16)
    ax43.set_xlabel(r'rho', fontsize=16)
    ax43.plot(beamData.debug['rho_nbeams'], beamData.debug['dvol'])

    plt.show()


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from math import sqrt, pi
import numpy as np

def calc_intrin_rot(M_orb, T_J, m):

    intrin_rot = 2 / sqrt(pi) * M_orb * np.sqrt(2 * T_J / m)
    return intrin_rot



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri May 18 15:04:43 2018

@author: max
"""
from __future__ import division
import numpy as np
from shapely.geometry import LineString
from shapely.ops import polygonize, linemerge
import sys
from contours.quad import QuadContourGenerator
from GT3.Core.Functions.Cut import cut


class Pfr:
    def __init__(self, inp, core):

        # R = inp.psirz_exp[:, 0].reshape(-1, 65)
        # Z = inp.psirz_exp[:, 1].reshape(-1, 65)

        self.pfr_lines(inp, core)
        self.pfr_nT(inp, core)

    def pfr_lines(self, inp, core):

        c = QuadContourGenerator.from_rectilinear(core.psi_data.R[0], core.psi_data.Z[:, 0], core.psi_data.psi_norm)
        contours = c.contour(0.999)

        if len(contours) == 1:
            # then we're definitely dealing with a surface inside the seperatrix
            raise ValueError("Did not find PFR flux surface. Stopping.")
        else:
            # we need to find the surface that is contained within the private flux region
            for j, contour in enumerate(contours):
                # if the contour is entirely below the x-point and contour extends to both the left and the
                # right of the x-point horizontally, then the contour is almost certainly the desired PFR contour
                if np.amax(contour[:, 1]) < core.pts.xpt[1] and \
                        np.amin(contour[:, 0]) < core.pts.xpt[0] < np.amax(contour[:, 0]):
                    # then it's a probably a pfr flux surface, might need to add additional checks later

                    # make sure the line goes from inboard to outboard
                    if contour[-1, 0] < contour[0, 0]:
                        contour = np.flipud(contour)

                    # find cut points
                    cut_pt_ib = LineString(contour).intersection(inp.wall_line)[0]
                    cut_pt_ob = LineString(contour).intersection(inp.wall_line)[1]

                    dist1 = LineString(contour).project(cut_pt_ib, normalized=True)
                    cutline_temp = cut(LineString(contour), dist1)[1]

                    # reverse line point order so we can reliably find the second intersection point
                    cutline_temp_rev = LineString(np.flipud(np.asarray(cutline_temp.coords)))

                    dist2 = cutline_temp_rev.project(cut_pt_ob, normalized=True)
                    cutline_final_rev = cut(cutline_temp_rev, dist2)[1]

                    # reverse again for final pfr flux line
                    pfr_flux_line = LineString(np.flipud(np.asarray(cutline_final_rev.coords)))

                    # add pfr_line intersection points on inboard side
                    # for some reason, union freaks out when I try to do inboard and outboard
                    # at the same time.
                    union = inp.wall_line.union(cut(LineString(contour), 0.5)[0])
                    result = [geom for geom in polygonize(union)][0]
                    inp.wall_line = LineString(result.exterior.coords)

                    # add pfr line intersection points on outboard side
                    union = inp.wall_line.union(cut(LineString(contour), 0.5)[1])
                    result = [geom for geom in polygonize(union)][0]
                    inp.wall_line = LineString(result.exterior.coords)

                    # cut out pfr section of wall line
                    wall_pts = np.asarray(inp.wall_line.xy).T

                    wall_start_pos = np.where((wall_pts == cut_pt_ob).all(axis=1))[0][0]
                    wall_line_rolled = LineString(np.roll(wall_pts, -wall_start_pos, axis=0))
                    wall_line_cut_pfr = cut(wall_line_rolled,
                                            wall_line_rolled.project(cut_pt_ib, normalized=True))[0]

                    # create LineString with pfr line and section of wall line
                    self.pfr_line = linemerge((pfr_flux_line, wall_line_cut_pfr))
                    break

    def pfr_nT(self, inp, core):

        pfr_pts = np.asarray(self.pfr_line.xy).T
        pfr_ni = np.zeros(len(pfr_pts)) + inp.pfr_ni_val
        pfr_ne = np.zeros(len(pfr_pts)) + inp.pfr_ne_val
        pfr_Ti = np.zeros(len(pfr_pts)) + inp.pfr_Ti_val
        pfr_Te = np.zeros(len(pfr_pts)) + inp.pfr_Te_val

        # TODO: May need to convert these temperatures to keV
        pts_ni_pfr = np.column_stack((pfr_pts, pfr_ni))
        pts_ne_pfr = np.column_stack((pfr_pts, pfr_ne))
        pts_Ti_pfr = np.column_stack((pfr_pts, pfr_Ti))
        pts_Te_pfr = np.column_stack((pfr_pts, pfr_Te))

        brnd_pfr_dict = {}
        brnd_pfr_dict['ni'] = pts_ni_pfr
        brnd_pfr_dict['ne'] = pts_ne_pfr
        brnd_pfr_dict['Ti'] = pts_Ti_pfr
        brnd_pfr_dict['Te'] = pts_Te_pfr

        # grid_x, grid_y = np.mgrid[1:2.5:500j, -1.5:1.5:500j]
        # ni_for_plot = griddata(self.ni_pts[:, :-1], self.ni_pts[:, -1], (grid_x, grid_y))
        # Ti_for_plot = griddata(self.Ti_kev_pts[:, :-1], self.Ti_kev_pts[:, -1], (grid_x, grid_y))
        # plt.contourf(grid_x, grid_y, np.log10(Ti_for_plot), 500)
        # plt.colorbar()
        # sys.exit()
        pass
